#!/usr/bin/env python3

# procedure to get start and end times from Level-0 fpacket files for HARP2
#  The times are extracted from the detector image headers.  The detector number 
#  (1, 2, 3 or All) is also determined
#  Reference: HARP2 Telemetry Manual

__version__ = '1.3.0_2023-01-17'

import numpy as np
import os
import copy
from l0info_utils import read_packet, read_apid, get_anc_packet_time, is_bad_packet

def is_bad_image(fpacket,packetLength,fh):
    if sum(abs(np.frombuffer(fpacket[22:packetLength-1],dtype=np.uint8)-255))==0:
        print("Invalid packet image data at byte %d"%fh.tell())
        return True
    else:
        return False

def is_bad_apid(apid,fh):
    if apid == -1:
        return False
    if (apid < 750) or (apid > 799):
        print("Invalid packet header at byte %d."%fh.tell())
        return True
    else:
        return False

def l0info_harp(args, fh, output):
    print("Running l0info_harp (version: %s) \n" % __version__)
    
    status = 0
    if args.verbose:
        print("Reading HARP2 science data file.")
        
    file_size = os.fstat(fh.fileno()).st_size
        
    # Skip CFE header
    fh.seek(64)
    bSPW = True
    
    # Find first science fpacket (APID 751)
    apid = 0
    packetLength = 8
    while (fh.tell() < file_size) and (apid != 751) and (packetLength > 7): 
        fpacket, packetLength = read_packet(fh, bSPW)
        apid = read_apid(fpacket)
        if is_bad_apid(apid,fh):  return 104
    
    if is_bad_packet(packetLength,fh):  return 104
    # this check is disabled according to #188 on gitlab, where it states that
    # "this is a normal consequence of the science data collection and download"
    # if is_bad_image(fpacket,packetLength,fh):  return 111     
        
    # If no science packets, rewind and get times from HKT packets (APID 750)
    if (packetLength==0):
        print("No science packets found in file")
        status = 110
        
        fh.seek(64)
        apid = 0
        packetLength = 8
        while (fh.tell() < file_size) and (apid != 750) and (packetLength > 7):
            fpacket, packetLength = read_packet(fh, bSPW)
            apid = read_apid(fpacket)
            
        if is_bad_packet(packetLength,fh):  return 104
        
        if fpacket:
            mpacket = fpacket
            try:
                stime = get_anc_packet_time(mpacket[6:12],0,'msec')
                print("start_time=%s" % stime.strftime('%Y-%m-%dT%H:%M:%S.%f'))
                if output:
                    output.write("start_time=%s\n" % stime.strftime('%Y-%m-%dT%H:%M:%S.%f'))
            except:
                return 120
            
            while fpacket and (packetLength > 7):
                fpacket, packetLength = read_packet(fh, bSPW)
                apid = read_apid(fpacket)
                if apid == 750:  mpacket = fpacket
            
            if is_bad_packet(packetLength,fh):  return 104
            etime = get_anc_packet_time(mpacket[6:12],0,'msec')
            print("etime=%s" % etime.strftime('%Y-%m-%dT%H:%M:%S.%f'))  # debug LH
            
        try:
            print("stop_time=%s" % etime.strftime('%Y-%m-%dT%H:%M:%S.%f'))
            if output:
                output.write("stop_time=%s\n" % etime.strftime('%Y-%m-%dT%H:%M:%S.%f'))
        except Exception as e:
            print(e)
            return 120
            
        return status
        
    # Get detector number from AP header
    detnum = (fpacket[12] & 48)>>4
      
      # Find an image header
    str_imhead = "HARP" # [72, 65, 82, 80]    # Header start mark ASCII "HARP"
    imhp = -1
    while (imhp == -1):
        str_packet = "".join([chr(x) for x in fpacket])
        if str_imhead in str_packet:
            imhp = str_packet.index(str_imhead)
        
        if imhp == -1: # If no header in current fpacket, find next science fpacket
            apid = 0
            packetLength = 8
            while (fh.tell() < file_size) and (apid != 751) and (packetLength > 7):
                fpacket, packetLength = read_packet(fh, bSPW)
                apid = read_apid(fpacket)
                if is_bad_apid(apid,fh):  return 101
            
            if is_bad_packet(packetLength,fh):  return 104
            # if is_bad_image(fpacket,packetLength,fh):  return 111          
            if packetLength==0:
                print('No image headers found in file.')                
                return 110
      
            # Check for more than one detector in file
            detn = (fpacket[12] & 48)>>4
            if detn != detnum: detnum = 999

    if args.verbose:
        print("Image pointer: %d"%imhp)
    
    # Get start time
    try:
        stime = get_anc_packet_time(fpacket[imhp+7:imhp+13],0,'msec')
        print("start_time=%s" % stime.strftime('%Y-%m-%dT%H:%M:%S.%f'))
        if output:
            output.write("start_time=%s\n" % stime.strftime('%Y-%m-%dT%H:%M:%S.%f'))
    except:
        return 120

    # Read to end of file
    epacket = fpacket
    # Look for next image header
    packetLength = 8
    last_pct = 0
    while fh.tell() < file_size:
        apid = 0

        # Find next science fpacket
        while (fh.tell() < file_size) and (apid != 751) and (packetLength > 7):
            fpacket, packetLength = read_packet(fh, bSPW)
            apid = read_apid(fpacket)
            if is_bad_apid(apid,fh):  return 104
        
        if is_bad_packet(packetLength,fh):  return 104
        # if is_bad_image(fpacket,packetLength,fh):  return 111            
        
        if fpacket:
            # Check for more than one detector in file
            detn = (fpacket[12] & 48)>>4
            if detn != detnum: detnum = 999
            # Look for image header
            imh = -1
            while (imh == -1) and fh.tell() < file_size:
                if args.verbose:
                    imh_pct = int(fh.tell()/file_size*100)
                    if imh_pct - last_pct > 9:
                        print("%d%% of packets read." % imh_pct)
                        last_pct = imh_pct
                str_packet = "".join([chr(x) for x in fpacket])
                if str_imhead in str_packet:
                    imh = str_packet.index(str_imhead)
                
                if (imh == -1): # If no header in current fpacket, find next science fpacket
                    apid = 0
                    while (fh.tell() < file_size) and (apid != 751) and (packetLength > 7):
                        fpacket, packetLength = read_packet(fh, bSPW)
                        apid = read_apid(fpacket)
                        if is_bad_apid(apid,fh):  return 104
                    if is_bad_packet(packetLength,fh):  return 104
                    # if is_bad_image(fpacket,packetLength,fh):  return 111                    
            if fpacket and (imh > -1):  # fh.tell() < file_size:
                imhp = imh
                if args.verbose:
                    print("Image pointer: %d"%imhp)
                epacket = fpacket
    
    if args.verbose:
        print("100% of packets read.")
    
    try:
        if imhp > -1:
            etime = get_anc_packet_time(epacket[imhp+7:imhp+13],0,'msec')
        else:
            etime = stime
        print("stop_time=%s" % etime.strftime('%Y-%m-%dT%H:%M:%S.%f'))
        if output:
            output.write("stop_time=%s\n" % etime.strftime('%Y-%m-%dT%H:%M:%S.%f'))
    except Exception as e:
        print(e)
        return 120
    
    if detnum == 999:
        print("detector=all")
        if output:
            output.write("detector=all\n")
    else:
        print("detector=%d"%detnum)
        if output:
            output.write("detector=%d\n"%detnum)
    
    return status

